@@ -17,7 +17,7 @@ from datetime import datetime
 
 from pybufrkit.errors import PyBufrKitError
 from pybufrkit.constants import BASE_DIR, NBITS_PER_BYTE, PARAMETER_TYPE_TEMPLATE_DATA
-from pybufrkit.tables import get_table_group
+from pybufrkit.tables import TableGroupCacheManager
 
 log = logging.getLogger(__file__)
 
@@ -27,7 +27,7 @@ class SectionParameter(object):
     This class represents a Parameter of a Bufr Section.
     """
 
-    def __init__(self, name, nbits, data_type, expected, as_property):
+    def __init__(self, name, nbits, data_type, expected, as_property, value=None):
         # type: (str, int, str, object, bool) -> None
         self.name = name
         self.nbits = nbits
@@ -39,7 +39,7 @@ class SectionParameter(object):
         self.expected = expected
         self.as_property = as_property
         self.parent = None  # the section that parameter belongs to
-        self.value = None
+        self.value = value
 
 
 class SectionNamespace(OrderedDict):
@@ -124,7 +124,7 @@ class BufrSection(object):
             else:
                 nbits_offset += parameter.nbits
         else:
-            raise RuntimeError('Parameter "{}" not found'.format(parameter_name))
+            raise PyBufrKitError('Parameter "{}" not found'.format(parameter_name))
 
 
 # An indicator for using the fallback default edition of a section
@@ -229,7 +229,7 @@ class SectionConfigurer(object):
         # Get the default config for the section if specific edition is not found
         return section_configs.get(section_edition, section_configs[DEFAULT_SECTION_EDITION])
 
-    def configure_section_with_values(self, bufr_message, section_index, values):
+    def configure_section_with_values(self, bufr_message, section_index, values, overrides=None):
         """
         Initialise and Configure a section for the give section index and version
         and also populate the value of each section parameter with the given list
@@ -246,7 +246,7 @@ class SectionConfigurer(object):
                 'Number of Section parameters ({}) not equal to number of values to be encoded ({})'.format(
                     len(section), len(values))
             for idx, parameter in enumerate(section):
-                parameter.value = values[idx]
+                parameter.value = values[idx] if overrides is None else overrides.get(parameter.name, values[idx])
 
         return section
 
@@ -316,9 +316,11 @@ class BufrMessage(object):
         self.table_group_key = None
         self._edition = None
         # Default to zero for edition 1 which does not have this parameter
-        self._master_table_number = 0
+        self._master_table_number = SectionParameter('master_table_number',
+                                                     0, 'unit', None, True, 0)
         # Default to zero for edition 1 and 2 which do not have this parameter
-        self._originating_subcentre = 0
+        self._originating_subcentre = SectionParameter('originating_subcentre',
+                                                       0, 'unit', None, True, 0)
 
     def add_section(self, section):
         """
@@ -338,7 +340,7 @@ class BufrMessage(object):
             one is not available.
         :return: A tuple of BufrTemplate and the associated TableGroup
         """
-        table_group = get_table_group(
+        table_group = TableGroupCacheManager.get_table_group(
             tables_root_dir=tables_root_dir,
             master_table_number=self.master_table_number.value,
             originating_centre=self.originating_centre.value,
@@ -504,6 +506,14 @@ class BufrMessage(object):
     def is_section2_presents(self, new_value):
         self._is_section2_presents = new_value
 
+    @property
+    def data_category(self):
+        return self._data_category
+
+    @data_category.setter
+    def data_category(self, new_value):
+        self._data_category = new_value
+
     @property
     def n_subsets(self):
         return self._n_subsets
