@@ -18,7 +18,7 @@ from pybufrkit.constants import (DEFAULT_TABLES_DIR,
                                  UNITS_CODE_TABLE,
                                  UNITS_FLAG_TABLE,
                                  UNITS_STRING)
-from pybufrkit.errors import UnknownDescriptor
+from pybufrkit.errors import PyBufrKitError, UnknownDescriptor
 from pybufrkit.bufr import SectionConfigurer
 from pybufrkit.descriptors import (ElementDescriptor,
                                    FixedReplicationDescriptor,
@@ -58,6 +58,11 @@ class AuditedList(list):
         log.debug('{!r}'.format(p_object))
         super(AuditedList, self).append(p_object)
 
+    def __getitem__(self, item):
+        value = super(AuditedList, self).__getitem__(item)
+        log.debug('{!r}'.format(value))
+        return value
+
 
 class CoderState(object):
     """
@@ -91,17 +96,20 @@ class CoderState(object):
 
         # For uncompressed data, the following two values will be changed during
         # subset context switching
-        self.decoded_descriptors = self.decoded_descriptors_all_subsets[0]
-        self.bitmap_links = self.bitmap_links_all_subsets[0]
+        self.decoded_descriptors = [] if n_subsets == 0 else self.decoded_descriptors_all_subsets[0]
+        self.bitmap_links = [] if n_subsets == 0 else self.bitmap_links_all_subsets[0]
 
         # When debug is turned on, use AuditedList for more logging messages.
         # Each element in the values all_subsets is different compressed or not
         if logging.root.level == logging.getLevelName('DEBUG'):
-            self.decoded_values_all_subsets = decoded_values_all_subsets or [AuditedList() for _ in range(n_subsets)]
+            if decoded_values_all_subsets:
+                self.decoded_values_all_subsets = [AuditedList(vals) for vals in decoded_values_all_subsets]
+            else:
+                self.decoded_values_all_subsets = [AuditedList() for _ in range(n_subsets)]
         else:
             self.decoded_values_all_subsets = decoded_values_all_subsets or [[] for _ in range(n_subsets)]
 
-        self.decoded_values = self.decoded_values_all_subsets[0]
+        self.decoded_values = [] if n_subsets == 0 else self.decoded_values_all_subsets[0]
 
         self.idx_value = 0  # only needed for encoder
 
@@ -184,7 +192,7 @@ class CoderState(object):
             value = self.decoded_values[idx]
 
         if value is None or value < 0:
-            raise RuntimeError('Delayed replication factor must be >= 0: ({!r})'.format(value))
+            raise PyBufrKitError('Delayed replication factor must be >= 0: got ({!r})'.format(value))
 
         return value
 
@@ -204,7 +212,7 @@ class CoderState(object):
                     if len(self.back_referenced_descriptors) == len(bitmap):
                         break
         if len(self.back_referenced_descriptors) != len(bitmap):
-            raise RuntimeError('Back referenced descriptors not matching defined Bitmap')
+            raise PyBufrKitError('Back referenced descriptors not matching defined Bitmap')
 
         # Lastly, get all the descriptors that has a corresponding Zero bit value
         self.bitmapped_descriptors = [
@@ -290,7 +298,7 @@ class Coder(object):
         for member in members:
             member_type = type(member)
 
-            log.debug('Processing {}'.format(member))
+            log.debug('Processing {} {}'.format(member, member.name if hasattr(member, 'name') else ''))
 
             # TODO: NOT using if-elif for following checks because they may co-exist???
             #      It is highly unlikely if not impossible
@@ -308,7 +316,8 @@ class Coder(object):
                         #       So it helps to keep the structure intact??
 
             # Currently defining new reference values
-            if state.nbits_of_new_refval:
+            # For ElementDescriptor only. This makes sense though not explicitly stated in the manual
+            if state.nbits_of_new_refval and member_type is ElementDescriptor:
                 self.process_define_new_refval(state, bit_operator, member)
                 continue
 
@@ -349,10 +358,10 @@ class Coder(object):
         :param bit_operator:
         :param descriptor:
         """
-        log.debug('Defining new reference value for '.format(descriptor))
+        log.debug('Defining new reference value for {}'.format(descriptor))
         if descriptor.unit == UNITS_STRING:
-            raise RuntimeError('Cannot define new reference value for descriptor of string value')
-        self.process_new_refval(bit_operator, descriptor, state.nbits_of_new_refval)
+            raise PyBufrKitError('Cannot define new reference value for descriptor of string value')
+        self.process_new_refval(state, bit_operator, descriptor, state.nbits_of_new_refval)
 
     def process_skipped_local_descriptor(self, state, bit_operator, descriptor):
         """
@@ -491,6 +500,7 @@ class Coder(object):
         if descriptor.id in (31011, 31012):
             raise NotImplementedError('delayed repetition descriptor')
 
+        log.debug('Processing {}'.format(descriptor.factor))
         self.process_element_descriptor(state, bit_operator, descriptor.factor)
         for _ in range(self.get_value_for_delayed_replication_factor(state)):
             self.process_members(state, bit_operator, descriptor.members)
@@ -512,11 +522,12 @@ class Coder(object):
             state.scale_offset = (operand_value - 128) if operand_value else 0
 
         elif operator_code == 203:  # new reference value
-            if operand_value == 0:
-                state.new_refvals = {}  # TODO: this seems not used
-                state.nbits_of_new_refval = operand_value
-            else:  # 255
+            if operand_value == 255:  # 255 is to conclude not cancel
                 state.nbits_of_new_refval = 0
+            else:
+                state.nbits_of_new_refval = operand_value
+                if operand_value == 0:
+                    state.new_refvals = {}
 
         elif operator_code == 204:  # associated field
             if operand_value == 0:
@@ -578,7 +589,7 @@ class Coder(object):
             self.process_constant(state, bit_operator, descriptor, 0)
 
         else:  # TODO: 241, 242, 243
-            raise NotImplemented('Operator Descriptor {} not implemented'.format(descriptor))
+            raise NotImplementedError('Operator Descriptor {} not implemented'.format(descriptor))
 
     def process_sequence_descriptor(self, state, bit_operator, descriptor):
         self.process_members(state, bit_operator, descriptor.members)
