@@ -17,8 +17,8 @@ from pybufrkit.constants import (BITPOS_START,
                                  NUMERIC_MISSING_VALUES,
                                  PARAMETER_TYPE_TEMPLATE_DATA,
                                  PARAMETER_TYPE_UNEXPANDED_DESCRIPTORS)
+from pybufrkit.errors import PyBufrKitError
 from pybufrkit.bitops import get_bit_writer
-from pybufrkit.utils import nested_json_to_flat_json
 from pybufrkit.bufr import BufrMessage
 from pybufrkit.templatedata import TemplateData
 from pybufrkit.descriptors import Descriptor
@@ -57,10 +57,17 @@ class Encoder(Coder):
                  definitions_dir=None,
                  tables_root_dir=None,
                  ignore_declared_length=True,
-                 compiled_template_cache_max=None):
+                 compiled_template_cache_max=None,
+                 master_table_number=None,
+                 master_table_version=None):
 
         super(Encoder, self).__init__(definitions_dir, tables_root_dir)
         self.ignore_declared_length = ignore_declared_length
+        self.overrides = {}
+        if master_table_number:
+            self.overrides['master_table_number'] = master_table_number
+        if master_table_version:
+            self.overrides['master_table_version'] = master_table_version
 
         # Only enable template compilation if cache is requested
         if compiled_template_cache_max is not None:
@@ -69,14 +76,13 @@ class Encoder(Coder):
         else:
             self.compiled_template_manager = None
 
-    def process(self, s, file_path='<string>', is_nested_json=False, wire_template_data=True):
+    def process(self, s, file_path='<string>', wire_template_data=True):
         """
         Entry point for the encoding process. The process encodes a JSON format
         message to BUFR message.
 
         :param s: A JSON or its string serialized form
         :param file_path: The file path to the JSON file.
-        :param is_nested_json: Whether the input string has the nested JSON format
         :param wire_template_data: Whether to wire the template data to construct
             a fully hierarchical structure from the flat lists.
 
@@ -85,13 +91,10 @@ class Encoder(Coder):
 
         if isinstance(s, (six.binary_type, six.text_type)):
             # TODO: ensure all strings are loaded as plain ascii instead of unicode from JSON
-            json_data = json.loads(s, encoding='latin-1')
+            json_data = json.loads(s) if six.PY3 else json.loads(s, encoding='latin-1')
         else:
             json_data = s
 
-        if is_nested_json:
-            json_data = nested_json_to_flat_json(json_data)
-
         bit_writer = get_bit_writer()
         bufr_message = BufrMessage(filename=file_path)
 
@@ -102,7 +105,8 @@ class Encoder(Coder):
         index_offset = 0
         while True:
             section = self.section_configurer.configure_section_with_values(
-                bufr_message, section_index, json_data[section_index - index_offset])
+                bufr_message, section_index, json_data[section_index - index_offset],
+                self.overrides)
             section_index += 1
             if section is None:  # optional section is not present
                 index_offset += 1  # This section should not be counted
@@ -123,7 +127,7 @@ class Encoder(Coder):
                 section.get_metadata(BITPOS_START) + section.get_parameter_offset('length')
             )
         elif bufr_message.length.value != nbytes_write:
-            raise RuntimeError('Write exceeds declared total length {} by {} bytes'.format(
+            raise PyBufrKitError('Write exceeds declared total length {} by {} bytes'.format(
                 bufr_message.length.value, nbytes_write - bufr_message.length.value
             ))
 
@@ -184,7 +188,7 @@ class Encoder(Coder):
                     log.debug('Padding {} bits to for declared length of the section'.format(nbits_unwrite))
                     bit_writer.skip(nbits_unwrite)
                 elif nbits_unwrite < 0:
-                    raise RuntimeError('Writing exceeds declared section length {} by {} bytes'.format(
+                    raise PyBufrKitError('Writing exceeds declared section length {} by {} bytes'.format(
                         section.section_length.value, -nbits_unwrite // NBITS_PER_BYTE
                     ))
 
@@ -434,7 +438,7 @@ class Encoder(Coder):
             for value in values:
                 bit_writer.write_uint(value - min_value, nbits_diff)
 
-    def process_numeric_with_new_refval(self, state, bit_writer,
+    def process_numeric_of_new_refval(self, state, bit_writer,
                                         descriptor, nbits, scale_powered,
                                         refval_factor):
         """
