@@ -21,6 +21,7 @@ available in the cache yet (and save them to the cache for future use).
 Template are then processed by Coder in conjunction with a bit operator to
 create a BufrMessage object.
 """
+from __future__ import print_function
 from __future__ import absolute_import
 import os
 import json
@@ -37,8 +38,9 @@ from pybufrkit.descriptors import (ElementDescriptor,
                                    FixedReplicationDescriptor, DelayedReplicationDescriptor,
                                    OperatorDescriptor, SequenceDescriptor, BufrTemplate,
                                    UndefinedElementDescriptor, UndefinedSequenceDescriptor)
+from pybufrkit.utils import generate_quiet
 
-__all__ = ['TableGroupKey', 'get_table_group', 'get_table_group_by_key']
+__all__ = ['TableGroupKey', 'TableGroupCacheManager']
 
 log = logging.getLogger(__file__)
 
@@ -53,7 +55,7 @@ MAXIMUM_NUMBER_OF_CACHED_TABLE_GROUPS = 50
 DEFAULT_MASTER_TABLE_NUMBER = 0
 DEFAULT_ORIGINATING_CENTRE = 0
 DEFAULT_ORIGINATING_SUBCENTRE = 0
-DEFAULT_MASTER_TABLE_VERSION = 27
+DEFAULT_MASTER_TABLE_VERSION = 33
 DEFAULT_LOCAL_TABLE_VERSION = 0
 
 
@@ -101,7 +103,7 @@ def normalize_tables_sn(tables_root_dir,
 
     # Ensure the master table number exists
     if not os.path.isdir(os.path.join(tables_root_dir, master_table_number_string)):
-        log.warn('Fallback to default master table number: {} ({} not found)'.format(
+        log.warning('Fallback to default master table number: {} ({} not found)'.format(
             DEFAULT_MASTER_TABLE_NUMBER, master_table_number_string
         ))
         master_table_number_string = str(DEFAULT_MASTER_TABLE_NUMBER)
@@ -115,7 +117,7 @@ def normalize_tables_sn(tables_root_dir,
                                   )):
         wmo_tables_sn = (master_table_number_string, centres, master_table_version_string)
     else:
-        log.warn('Fallback to default master table version {} ({} not found)'.format(
+        log.warning('Fallback to default master table version {} ({} not found)'.format(
             DEFAULT_MASTER_TABLE_VERSION, master_table_version_string
         ))
         wmo_tables_sn = (master_table_number_string, centres, str(DEFAULT_MASTER_TABLE_VERSION))
@@ -133,12 +135,12 @@ def normalize_tables_sn(tables_root_dir,
                                           local_table_version_string)):
                 local_tables_sn = (master_table_number_string, centres, local_table_version_string)
                 if idx != 0:
-                    log.warn('Fallback to default local sub-centre {} ({} not found)'.format(
+                    log.warning('Fallback to default local sub-centre {} ({} not found)'.format(
                         DEFAULT_ORIGINATING_SUBCENTRE, originating_subcentre
                     ))
                 break
         else:
-            log.warn('Cannot find sub-centre {} nor valid default. Local table not in use.'.format(
+            log.warning('Cannot find sub-centre {} nor valid default. Local table not in use.'.format(
                 originating_subcentre
             ))
             local_tables_sn = None
@@ -150,22 +152,19 @@ def normalize_tables_sn(tables_root_dir,
 
 
 class BaseTable(object):
-    def __init__(self, tables_root_dir, wmo_tables_sn, local_tables_sn):
-        self.tables_root_dir, self.wmo_tables_sn, self.local_tables_sn = (
-            tables_root_dir, wmo_tables_sn, local_tables_sn
-        )
+    def __init__(self, table_group_key, extra_entries=()):
+        self.table_group_key = table_group_key
+        self.extra_entries = extra_entries
 
     def __eq__(self, other):
-        return (
-            type(self) is type(other) and
-            self.tables_root_dir == other.tables_root_dir and
-            self.wmo_tables_sn == other.wmo_tables_sn and
-            self.local_tables_sn == other.local_tables_sn
-        )
+        return type(self) is type(other) and self.table_group_key == other.table_group_key
 
     def __str__(self):
         return '{}: {} - {}, {}'.format(
-            self.__class__.__name__, self.tables_root_dir, self.wmo_tables_sn, self.local_tables_sn
+            self.__class__.__name__,
+            self.table_group_key.tables_root_dir,
+            self.table_group_key.wmo_tables_sn,
+            self.table_group_key.local_tables_sn
         )
 
     @property
@@ -174,13 +173,13 @@ class BaseTable(object):
 
     @property
     def tables_dir_wmo(self):
-        return os.path.join(os.path.join(self.tables_root_dir, *self.wmo_tables_sn))
+        return os.path.join(os.path.join(self.table_group_key.tables_root_dir, *self.table_group_key.wmo_tables_sn))
 
     @property
     def tables_dir_local(self):
         return (
-            os.path.join(os.path.join(self.tables_root_dir, *self.local_tables_sn))
-            if self.local_tables_sn else None
+            os.path.join(os.path.join(self.table_group_key.tables_root_dir, *self.table_group_key.local_tables_sn))
+            if self.table_group_key.local_tables_sn else None
         )
 
     def load_json_files(self, fname):
@@ -191,6 +190,9 @@ class BaseTable(object):
             with open(os.path.join(tables_dir, fname)) as ins:
                 contents.append(json.load(ins))
 
+        if self.extra_entries:
+            contents.append(self.extra_entries)
+
         return contents
 
 
@@ -291,12 +293,13 @@ def _descriptors_from_ids_iter(b, c, r, d, next_id):
             if isinstance(descriptor, DelayedReplicationDescriptor):
                 descriptor.factor = b.lookup(next_id())
 
-            g = (next_id() for _ in range(descriptor.n_items))
+            g = generate_quiet(range(descriptor.n_items), next_id)
             # TODO: check whether the actual number of members equals to n_items
             descriptor.members = _descriptors_from_ids_iter(b, c, r, d, functools.partial(next, g))
             descriptors.append(descriptor)
 
         else:  # element descriptor
+            # TODO: if this descriptor is preceded by a 206XXX, it should turn into skipped local
             descriptors.append(b.lookup(id_))
 
     return descriptors
@@ -354,12 +357,12 @@ class BufrTableGroup(_BufrTableGroup):
 
     def __eq__(self, other):
         return (
-            isinstance(other, BufrTableGroup) and
-            self.A == other.A and
-            self.B == other.B and
-            self.C == other.C and
-            self.D == other.D and
-            self.R == other.R
+                isinstance(other, BufrTableGroup) and
+                self.A == other.A and
+                self.B == other.B and
+                self.C == other.C and
+                self.D == other.D and
+                self.R == other.R
         )
 
     def __str__(self):
@@ -375,7 +378,7 @@ class BufrTableGroup(_BufrTableGroup):
 
     @property
     def key(self):
-        return TableGroupKey(self.A.tables_root_dir, self.A.wmo_tables_sn, self.A.local_tables_sn)
+        return self.A.table_group_key
 
     def lookup(self, id_):
         """
@@ -413,7 +416,41 @@ class BufrTableGroup(_BufrTableGroup):
         """
         Build BUFR Template from a list of IDs
         """
-        return BufrTemplate(members=self.descriptors_from_ids(*ids))
+        members = self.descriptors_from_ids(*ids)
+        if TableGroupCacheManager.has_extra_entries():
+            members = _fix_ncep_descriptors(members)
+        bufr_template = BufrTemplate(members=members)
+        return bufr_template
+
+
+def _fix_ncep_descriptors(descriptors):
+    """
+    Fix ill-defined sequence descriptor, e.g. NCEP 360001, which is consisted of
+    101000, 031002. Note that the actual descriptor to replicated is not part of
+    the sequence descriptor! This is pure evil!!
+    """
+    from copy import deepcopy
+    ret = []
+    while len(descriptors) > 0:
+        descriptor = deepcopy(descriptors.pop(0))
+        if isinstance(descriptor, SequenceDescriptor):
+            if (len(descriptor.members) == 1 and
+                    isinstance(descriptor.members[0], (FixedReplicationDescriptor, DelayedReplicationDescriptor)) and
+                    len(descriptor.members[0].members) == 0):
+                descriptors.insert(0, descriptor.members[0])
+            else:
+                descriptor.members = _fix_ncep_descriptors(descriptor.members)
+                ret.append(descriptor)
+        elif isinstance(descriptor, (FixedReplicationDescriptor, DelayedReplicationDescriptor)):
+            if len(descriptor.members) == 0:
+                assert descriptor.n_items == 1, 'Fix for replication descriptor expects 1 member, got {}'.format(
+                    len(descriptor.members))
+                descriptor.members = [descriptors.pop(0)]
+            descriptor.members = _fix_ncep_descriptors(descriptor.members)
+            ret.append(descriptor)
+        else:
+            ret.append(descriptor)
+    return ret
 
 
 class TableGroupCache(object):
@@ -427,6 +464,8 @@ class TableGroupCache(object):
     def __init__(self):
         # TODO: set maximum number of table groups to be cached
         self._groups = {}
+        self.extra_b_entries = {}
+        self.extra_d_entries = {}
 
     def get(self, table_group_key):
         if table_group_key not in self._groups:
@@ -435,74 +474,98 @@ class TableGroupCache(object):
                 for _ in range(len(self._groups) + 1 - MAXIMUM_NUMBER_OF_CACHED_TABLE_GROUPS):
                     self._groups.popitem()
 
-            a = TableA(*table_group_key)
-            b = TableB(*table_group_key)
-            c = TableC(*table_group_key)
-            r = TableR(*table_group_key)
-            d = TableD(b, c, r, *table_group_key)
+            a = TableA(table_group_key)
+            b = TableB(table_group_key, self.extra_b_entries)
+            c = TableC(table_group_key)
+            r = TableR(table_group_key)
+            d = TableD(b, c, r, table_group_key, self.extra_d_entries)
             self._groups[table_group_key] = BufrTableGroup(a, b, c, d, r)
 
         return self._groups[table_group_key]
 
+    def has_extra_entries(self):
+        return self.extra_b_entries or self.extra_d_entries
 
-_TABLE_GROUP_CACHE = TableGroupCache()
+    def invalidate(self):
+        self._groups.clear()
+
+    def add_extra_entries(self, b_entries, d_entries):
+        self.extra_b_entries.update(b_entries)
+        self.extra_d_entries.update(d_entries)
 
 
-def get_table_group(tables_root_dir=None,
-                    master_table_number=None,
-                    originating_centre=None,
-                    originating_subcentre=None,
-                    master_table_version=None,
-                    local_table_version=None,
-                    normalize=True):
-    """
-    Retrieve a Table Group using given information that can be converted
-    to a Table Group Key.
+class TableGroupCacheManager(object):
+    _TABLE_GROUP_CACHE = TableGroupCache()
 
-    :param master_table_number:
-    :param originating_centre:
-    :param originating_subcentre:
-    :param master_table_version:
-    :param local_table_version:
-    :param str tables_root_dir: Root directory to read the BUFR tables
-    :param int|bool normalize: Whether the program tries to fix non-exist tables SN by
-                      using default values. This is generally useful for
-                      decoding. But could be misleading when encoding.
-    :rtype: BufrTableGroup
-    """
+    @classmethod
+    def has_extra_entries(cls):
+        return cls._TABLE_GROUP_CACHE.has_extra_entries()
 
-    tables_root_dir = tables_root_dir or DEFAULT_TABLES_DIR
+    @classmethod
+    def invalidate(cls):
+        cls._TABLE_GROUP_CACHE.invalidate()
 
-    if normalize:
-        wmo_tables_sn, local_tables_sn = normalize_tables_sn(
-            tables_root_dir,
-            master_table_number or DEFAULT_MASTER_TABLE_NUMBER,
-            originating_centre or DEFAULT_ORIGINATING_CENTRE,
-            originating_subcentre or DEFAULT_ORIGINATING_SUBCENTRE,
-            master_table_version or DEFAULT_MASTER_TABLE_VERSION,
-            local_table_version or DEFAULT_LOCAL_TABLE_VERSION
-        )
-    else:
-        wmo_tables_sn, local_tables_sn = get_tables_sn(
-            master_table_number,
-            originating_centre,
-            originating_subcentre,
-            master_table_version,
-            local_table_version
-        )
+    @classmethod
+    def add_extra_entries(cls, b_entries, d_entries):
+        cls._TABLE_GROUP_CACHE.add_extra_entries(b_entries, d_entries)
 
-    table_group_key = TableGroupKey(tables_root_dir, wmo_tables_sn, local_tables_sn)
+    @classmethod
+    def get_table_group_by_key(cls, table_group_key):
+        """
+        Retrieve a Table Group with the given key.
 
-    # TODO: catch error on file reading?
-    return get_table_group_by_key(table_group_key)
+        :param TableGroupKey table_group_key: The key for a table group.
+        :rtype: BufrTableGroup
+        """
+        log.debug('Reading table group of key: {}'.format(table_group_key))
+        return cls._TABLE_GROUP_CACHE.get(table_group_key)
 
+    @classmethod
+    def get_table_group(cls,
+                        tables_root_dir=None,
+                        master_table_number=None,
+                        originating_centre=None,
+                        originating_subcentre=None,
+                        master_table_version=None,
+                        local_table_version=None,
+                        normalize=True):
+        """
+        Retrieve a Table Group using given information that can be converted
+        to a Table Group Key.
 
-def get_table_group_by_key(table_group_key):
-    """
-    Retrieve a Table Group with the given key.
+        :param master_table_number:
+        :param originating_centre:
+        :param originating_subcentre:
+        :param master_table_version:
+        :param local_table_version:
+        :param str tables_root_dir: Root directory to read the BUFR tables
+        :param int|bool normalize: Whether the program tries to fix non-exist tables SN by
+                          using default values. This is generally useful for
+                          decoding. But could be misleading when encoding.
+        :rtype: BufrTableGroup
+        """
 
-    :param TableGroupKey table_group_key: The key for a table group.
-    :rtype: BufrTableGroup
-    """
-    log.debug('Reading table group of key: {}'.format(table_group_key))
-    return _TABLE_GROUP_CACHE.get(table_group_key)
+        tables_root_dir = tables_root_dir or DEFAULT_TABLES_DIR
+
+        if normalize:
+            wmo_tables_sn, local_tables_sn = normalize_tables_sn(
+                tables_root_dir,
+                master_table_number or DEFAULT_MASTER_TABLE_NUMBER,
+                originating_centre or DEFAULT_ORIGINATING_CENTRE,
+                originating_subcentre or DEFAULT_ORIGINATING_SUBCENTRE,
+                master_table_version or DEFAULT_MASTER_TABLE_VERSION,
+                local_table_version or DEFAULT_LOCAL_TABLE_VERSION
+            )
+        else:
+            wmo_tables_sn, local_tables_sn = get_tables_sn(
+                master_table_number,
+                originating_centre,
+                originating_subcentre,
+                master_table_version,
+                local_table_version
+            )
+
+        table_group_key = TableGroupKey(tables_root_dir, wmo_tables_sn, local_tables_sn)
+
+        # TODO: catch error on file reading?
+        return TableGroupCacheManager.get_table_group_by_key(table_group_key)
